#ifndef CORE_UTILITY_HPP
#define CORE_UTILITY_HPP

#include "core/Meta.hpp"
#include "core/Types.hpp"

size_t popCount(u64 u);

template<typename T>
inline T interpolate(const T& a, const T& b, float factor) {
    return a * (1.0f - factor) + b * factor;
}

static constexpr float PI = 3.14159265358979323846f;

template<typename T>
inline T radianToDegree(const T& radians) {
    return radians * static_cast<T>(180.0f / PI);
}

template<typename T>
inline T degreeToRadian(const T& radians) {
    return radians * static_cast<T>(PI / 180.0f);
}

template<typename T>
inline T max(const T& a, const T& b) {
    return a > b ? a : b;
}

template<typename T>
inline T min(const T& a, const T& b) {
    return a < b ? a : b;
}

template<typename T>
inline T clamp(const T& t, const T& from, const T& to) {
    return max(min(t, to), from);
}

using ExitHandler = void (*)(int, void*);
[[noreturn]] void exitWithHandler(const char* file, int line, int value);
void setExitHandler(ExitHandler h, void* data);
#define EXIT(exitValue) exitWithHandler(__FILE__, __LINE__, exitValue)

using OutOfMemoryHandler = void (*)(void*);
void setOutOfMemoryHandler(OutOfMemoryHandler h, void* data);

#ifdef CHECK_MEMORY
void* coreDebugAllocate(const char* file, int line, size_t n);
void* coreDebugZeroAllocate(const char* file, int line, size_t n);
void* coreDebugReallocate(const char* file, int line, void* p, size_t n);
void coreFreeDebug(void* p);
void printMemoryReport(void);
#define coreAllocate(n) coreDebugAllocate(__FILE__, __LINE__, n)
#define coreZeroAllocate(n) coreDebugZeroAllocate(__FILE__, __LINE__, n)
#define coreReallocate(p, n) coreDebugReallocate(__FILE__, __LINE__, p, n)
#define coreFree(p) coreFreeDebug(p)
void* operator new(size_t count, const char* file, int line);
void* operator new[](size_t count, const char* file, int line);
#define coreNew(type, ...) new(__FILE__, __LINE__) type(__VA_ARGS__)
#define coreNewN(type, n) new(__FILE__, __LINE__) type[n]
#define coreDelete(p) delete(p)
#define coreDeleteN(p) delete[](p)
#else
void* coreAllocate(size_t n);
void* coreZeroAllocate(size_t n);
void* coreReallocate(void* p, size_t n);
void coreFree(void* p);
#define printMemoryReport()
#define coreNew(type, ...) new type(__VA_ARGS__)
#define coreNewN(type, n) new type[n]
#define coreDelete(p) delete(p)
#define coreDeleteN(p) delete[](p)
#endif

bool sleepMillis(i64 millis);
bool sleepNanos(i64 nanos);
i64 getNanos(void);

template<typename T>
void bubbleSort(T* data, size_t n) {
    bool swapped = true;
    while(swapped && n > 0) {
        swapped = false;
        n--;
        for(size_t i = 0; i < n; i++) {
            if(data[i] > data[i + 1]) {
                Core::swap(data[i], data[i + 1]);
                swapped = true;
            }
        }
    }
}

#endif
